#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom Tensorflow losses.
"""
from multiprocessing import Pool, cpu_count
import numpy as np

import tensorflow as tf
tf.keras.backend.set_floatx('float64')
tf.config.threading.set_inter_op_parallelism_threads(2)
tf.config.threading.set_intra_op_parallelism_threads(8)
tf.config.set_soft_device_placement(True)
from tensorflow.python.ops import math_ops
from tensorflow.keras.losses import Loss

from .ops import rectified_exp, d_rectified_exp


class MINDLoss(Loss):  
	'''
	MIND loss function: :math:`-E_P(T(x, y)^T\theta) + \log E_Q(e^{T(x, y)^T\theta})`.
	'''
	def call(self, y_true, y_pred):
		''' '''
		p_samples = y_pred[:, 0]
		q_samples = y_pred[:, 1]
		mi = -tf.reduce_mean(p_samples) + math_ops.log(tf.reduce_mean(math_ops.exp(q_samples)))
		return mi


class ApproximateMINDLoss(Loss):
	'''
	MIND loss function with a gentler version of the exponential: :math:`-E_P(r_exp(T(x, y)^T\theta)) + \log E_Q(dr_exp(T(x, y)^T\theta)`. :math:`r_exp(t) = exp(t)` if :math:`t<0` and :math:`r_exp(t) = 1+x+(1/2)x^2+(1/6)x^2`.
	'''
	def call(self, y_true, y_pred):
		''' '''
		p_samples = y_pred[:, 0]
		q_samples = y_pred[:, 1]
		mi = -tf.reduce_mean(p_samples) + math_ops.log(tf.reduce_mean(rectified_exp(q_samples)))
		return mi


class RectifiedMINDLoss(Loss):
	'''
	Rectified-MIND loss function: :math:`-E_P(\log dr_exp((T(x, y)^T\theta)) + \log E_Q(dr_exp(T(x, y)^T\theta)`. :math:`r_exp(t) = exp(t)` if :math:`t<0` and :math:`r_exp(t) = 1+x+(1/2)x^2+(1/6)x^2`.
	'''
	def call(self, y_true, y_pred):
		''' '''
		p_samples = y_pred[:, 0]
		q_samples = y_pred[:, 1]
		mi = -tf.reduce_mean(math_ops.log(d_rectified_exp(p_samples))) + math_ops.log(tf.reduce_mean(d_rectified_exp(q_samples)))
		return mi






#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom math operations.
"""
from multiprocessing import Pool, cpu_count
import numpy as np

import tensorflow as tf
tf.keras.backend.set_floatx('float64')
tf.config.threading.set_inter_op_parallelism_threads(2)
tf.config.threading.set_intra_op_parallelism_threads(8)
tf.config.set_soft_device_placement(True)
from tensorflow.python.ops import math_ops

def rectified_exp(t):
	'''
	:math:`r_exp(t) = exp(t)` if :math:`t<0` and :math:`r_exp(t) = 1+x+(1/2)x^2+(1/6)x^3`.
	'''
	exp = math_ops.exp(t)
	approx_exp = 1.+t+(1./2.)*tf.math.pow(t, 2.)+(1./6.)*tf.math.pow(t, 3.)
	condition = tf.greater(t, 0.0)
	r_exp = tf.where(condition, x=approx_exp, y=exp)
	return r_exp


def d_rectified_exp(t):
	'''
	:math:`dr_exp(t) = exp(t)` if :math:`t<0` and :math:`dr_exp(t) = 1+x+(1/2)x^2`.
	'''
	dexp = math_ops.exp(t)
	approx_dexp = 1.+t+(1./2.)*tf.math.pow(t, 2.)
	condition = tf.greater(t, 0.0)
	dr_exp = tf.where(condition, x=approx_dexp, y=dexp)
	return dr_exp


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tensorflow learners.
"""
import numpy as np

import tensorflow as tf
from tensorflow.keras.callbacks import EarlyStopping, TerminateOnNaN
from tensorflow.keras.optimizers import Adam

from .generators import CopulaBatchGenerator, PFSBatchGenerator
from .models import CopulaModel, PFSModel, PFSOneShotModel
from .losses import MINDLoss, ApproximateMINDLoss, RectifiedMINDLoss


class CopulaLearner(object):
	'''
	Maximum-entropy learner.
	'''
	def __init__(self, d, beta_1=0.9, beta_2=0.999, epsilon=1e-07, amsgrad=False, \
			name='Adam', lr=0.01, subsets=[]):
		self.d = d
		self.model = CopulaModel(self.d, subsets=subsets)
		self.opt = Adam(beta_1=beta_1, beta_2=beta_2, epsilon=epsilon, amsgrad=amsgrad, \
			name=name, lr=lr)
		self.loss = MINDLoss()
		self.model.compile(optimizer=self.opt, loss=self.loss)
		self.copula_entropy = None


	def fit(self, z, batch_size=10000, steps_per_epoch=1000, epochs=20):
		''' '''
		z_gen = CopulaBatchGenerator(z, batch_size=batch_size, steps_per_epoch=steps_per_epoch)
		self.model.fit(z_gen, epochs=epochs, batch_size=batch_size, steps_per_epoch=steps_per_epoch, \
			callbacks=[EarlyStopping(patience=3, monitor='loss'), TerminateOnNaN()])
		self.copula_entropy = self.model.evaluate(z_gen)




class PFSLearner(object):
	'''
	Principal Feature Learner.
	'''
	def __init__(self, dx, dy=1, dox=0, doy=0, beta_1=0.9, beta_2=0.999, epsilon=1e-07, amsgrad=False, \
			name='Adam', lr=0.01):
		x_ixs = [_ for _ in range(dx)]
		y_ixs = [dx+_ for _ in range(dy)]
		ox_ixs = [dx+dy+_ for _ in range(dox)]
		oy_ixs = [dx+dy+dox+_ for _ in range(doy)]

		self.model = PFSModel(x_ixs, y_ixs, ox_ixs=ox_ixs, oy_ixs=oy_ixs)
		self.opt = Adam(beta_1=beta_1, beta_2=beta_2, epsilon=epsilon, amsgrad=amsgrad, \
			name=name, lr=lr)
		self.loss = RectifiedMINDLoss() # MINDLoss()
		self.model.compile(optimizer=self.opt, loss=self.loss)
		self.mutual_information = None
		self.feature_direction = None
		self.statistics = None


	def fit(self, x, y, ox=None, oy=None, batch_size=500, n_shuffle=5, epochs=20, mi_eps=0.00001):
		''' '''
		n = x.shape[0]
		steps_per_epoch = n//batch_size
		steps_per_epoch = min(max(steps_per_epoch, 100), 1000)

		z_gen = PFSBatchGenerator(x, y, ox=ox, oy=oy, batch_size=batch_size, \
			steps_per_epoch=steps_per_epoch, n_shuffle=n_shuffle)
		self.model.fit(z_gen, epochs=epochs, batch_size=batch_size, steps_per_epoch=steps_per_epoch, \
			callbacks=[EarlyStopping(patience=3, monitor='loss'), TerminateOnNaN()])
		self.mutual_information = -self.model.evaluate(z_gen)
		w = self.model.w_layer.get_weights()[0]

		if self.mutual_information < mi_eps:
			# Retrain to avoid MI collapse to 0.
			batch_size = 2*batch_size
			z_gen = PFSBatchGenerator(x, y, ox=ox, oy=oy, batch_size=batch_size, \
				steps_per_epoch=steps_per_epoch, n_shuffle=n_shuffle)
			self.model.fit(z_gen, epochs=epochs, batch_size=batch_size, steps_per_epoch=steps_per_epoch, \
				callbacks=[EarlyStopping(patience=3, monitor='loss'), TerminateOnNaN()])
			self.mutual_information = -self.model.evaluate(z_gen)
			w = self.model.w_layer.get_weights()[0]

		# The feature direction should be normal
		w = w.flatten()
		w = w/np.sqrt(np.dot(w, w))

		# The principal feature should point in the same direction as the target (i.e. <y, w^Tx> = cov(y, w^Tx) > 0)
		corr_sign = np.sign(np.corrcoef(y, np.dot(x, w))[0, 1])
		w = corr_sign*w

		self.feature_direction = w
		self.fx = self.model.fx(tf.constant(z_gen.z)).numpy()
		self.gy = self.model.gy(tf.constant(z_gen.z)).numpy()




class PFSOneShotLearner(object):
	'''
	Principal Feature Learner learning multiple principal features simultaneously.
	'''
	def __init__(self, dx, dy=1, beta_1=0.9, beta_2=0.999, epsilon=1e-07, amsgrad=False, \
			name='Adam', lr=0.01, p=1):
		x_ixs = [_ for _ in range(dx)]
		y_ixs = [dx+_ for _ in range(dy)]

		self.model = PFSOneShotModel(x_ixs, y_ixs, p=p)
		self.opt = Adam(beta_1=beta_1, beta_2=beta_2, epsilon=epsilon, amsgrad=amsgrad, \
			name=name, lr=lr)
		self.loss = RectifiedMINDLoss() # MINDLoss()
		self.model.compile(optimizer=self.opt, loss=self.loss)
		self.mutual_information = None
		self.feature_direction = None
		self.statistics = None


	def fit(self, x, y, batch_size=500, n_shuffle=5, epochs=20, mi_eps=0.00001):
		''' '''
		n = x.shape[0]
		steps_per_epoch = n//batch_size
		steps_per_epoch = min(max(steps_per_epoch, 100), 1000)

		z_gen = PFSBatchGenerator(x, y, batch_size=batch_size, \
			steps_per_epoch=steps_per_epoch, n_shuffle=n_shuffle)
		self.model.fit(z_gen, epochs=epochs, batch_size=batch_size, steps_per_epoch=steps_per_epoch)
		self.mutual_information = -self.model.evaluate(z_gen)
		w = self.model.w_layer.get_weights()[0]

		if self.mutual_information < mi_eps:
			# Retrain to avoid MI collapse to 0.
			batch_size = 2*batch_size
			z_gen = PFSBatchGenerator(x, y, batch_size=batch_size, \
				steps_per_epoch=steps_per_epoch, n_shuffle=n_shuffle)
			self.model.fit(z_gen, epochs=epochs, batch_size=batch_size, steps_per_epoch=steps_per_epoch)
			self.mutual_information = -self.model.evaluate(z_gen)
			w = self.model.w_layer.get_weights()[0]

		# The feature direction should be normal
		# This should already have been taken care of downstream, but just in case.
		w = w/np.linalg.norm(w, axis=0)

		# Each principal feature should point in the same direction as the target (i.e. <y, w^Tx> = cov(y, w^Tx) > 0)
		for j in range(w.shape[1]):
			corr_sign = np.sign(np.corrcoef(y.flatten(), np.dot(x, w[:, j]))[0, 1])
			w[:, j] = corr_sign*w[:, j]

		w = w.T

		self.feature_directions = w








#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2022 KXY TECHNOLOGIES, INC.
Author: Dr Yves-Laurent Kom Samo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from .improvability import *
from .model_explanation import *


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom Tensorflow generators.
"""
import numpy as np
import tensorflow as tf
tf.keras.backend.set_floatx('float64')
tf.config.threading.set_inter_op_parallelism_threads(2)
tf.config.threading.set_intra_op_parallelism_threads(8)
tf.config.set_soft_device_placement(True)
from tensorflow.keras.utils import Sequence

rankdata = lambda x: 1.+np.argsort(np.argsort(x, axis=0), axis=0)
class CopulaBatchGenerator(Sequence):
	''' 
	Random batch generator of maximum-entropy copula learning.
	'''
	def __init__(self, z, batch_size=1000, steps_per_epoch=100):
		self.batch_size = batch_size
		self.d = z.shape[1]
		self.n = z.shape[0]
		self.z = z
		self.steps_per_epoch = steps_per_epoch
		self.emp_u = rankdata(self.z)/(self.n + 1.)
		self.emp_u[np.isnan(self.z)] = 0.5

		if self.n < 200*self.d:
			dn = 200*self.d - self.n
			selected_rows = np.random.choice(self.n, dn, replace=True)
			emp_u = self.emp_u[selected_rows, :].copy()
			scale = 1./(100.*self.n)
			emp_u += (scale*np.random.rand(*emp_u.shape) - 0.5*scale)
			self.emp_u = np.concatenate([self.emp_u, emp_u], axis=0)
			self.n = self.emp_u.shape[0]

		self.batch_selector = np.random.choice(self.n, self.batch_size*self.steps_per_epoch, replace=True)
		self.batch_selector = self.batch_selector.reshape((self.steps_per_epoch, self.batch_size))


	def getitem_ndarray(self, idx):
		''' '''
		i = idx % self.steps_per_epoch
		selected_rows = self.batch_selector[i]
		emp_u_ = self.emp_u[selected_rows, :]
		z_p = emp_u_.copy()
		z_q = np.random.rand(*emp_u_.shape)

		z = np.empty((self.batch_size, self.d, 2))
		z[:, :, 0] = z_p
		z[:, :, 1] = z_q
		batch_x = z
		batch_y = np.ones((self.batch_size, 2))  # Not used  
		return batch_x, batch_y


	def __getitem__(self, idx):
		''' '''
		batch_x, batch_y = self.getitem_ndarray(idx)
		return tf.convert_to_tensor(batch_x), tf.convert_to_tensor(batch_y)


	def __len__(self):
		return self.steps_per_epoch



class PFSBatchGenerator(Sequence):
	'''
	Random batch generator.
	'''
	def __init__(self, x, y, ox=None, oy=None, batch_size=1000, steps_per_epoch=100, n_shuffle=5):
		assert x.shape[0] == y.shape[0]
		self.batch_size = batch_size
		self.n_shuffle = n_shuffle
		self.n = x.shape[0]

		x = x if len(x.shape) > 1 else x[:, None]
		y = y if len(y.shape) > 1 else y[:, None]
		ox = ox if ox is None or len(ox.shape) > 1 else ox[:, None]
		oy = oy if oy is None or len(oy.shape) > 1 else oy[:, None]

		self.x = x
		self.y = y
		self.ox = ox
		self.oy = oy
		self.z = np.concatenate([self.x, self.y, self.ox, self.oy], axis=1) if not self.ox is None else np.concatenate([self.x, self.y], axis=1)
		self.d = self.z.shape[1]

		self.steps_per_epoch = steps_per_epoch
		replace = False if self.n > self.batch_size*self.steps_per_epoch else True
		self.batch_selector = np.random.choice(self.n, self.batch_size*self.steps_per_epoch, replace=replace)
		self.batch_selector = self.batch_selector.reshape((self.steps_per_epoch, self.batch_size))
		
		
	def getitem_ndarray(self, idx):
		''' '''
		i = idx % self.steps_per_epoch
		selected_rows = self.batch_selector[i]
		x_ = self.x[selected_rows, :]
		y_ = self.y[selected_rows, :]
		z_ = self.z[selected_rows, :]
		if not self.ox is None:
			ox_ = self.ox[selected_rows, :]
			oy_ = self.oy[selected_rows, :]

		z_p = None
		z_q = None
		for _ in range(self.n_shuffle):
			z_p = z_.copy() if z_p is None else np.concatenate([z_p, z_.copy()], axis=0)
			y_q = y_.copy()
			randomize = np.arange(y_q.shape[0])
			np.random.shuffle(randomize)
			y_q = y_q[randomize]
			if not self.ox is None:
				oy_q = oy_.copy()
				oy_q = oy_q[randomize]
			z_q_ = np.concatenate([x_, y_q.copy(), ox_, oy_q], axis=1) if not self.ox is None else np.concatenate([x_, y_q.copy()], axis=1)
			z_q = z_q_.copy() if z_q is None else np.concatenate([z_q, z_q_.copy()], axis=0)

		z = np.empty((self.batch_size*self.n_shuffle, self.d, 2))
		z[:, :, 0] = z_p
		z[:, :, 1] = z_q
		batch_x = z
		batch_y = np.ones((self.batch_size*self.n_shuffle, 2))  # Not used   
		return batch_x, batch_y


	def __getitem__(self, idx):
		''' '''
		batch_x, batch_y = self.getitem_ndarray(idx)
		return tf.convert_to_tensor(batch_x), tf.convert_to_tensor(batch_y)
	
	def __len__(self):
		return self.steps_per_epoch







#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
TensorFlow Implementation of MIND ([1]) under Spearman rank correlation constraints.

[1] Kom Samo, Y. (2021). Inductive Mutual Information Estimation: A Convex Maximum-Entropy Copula Approach . <i>Proceedings of The 24th International Conference on Artificial Intelligence and Statistics</i>, in <i>Proceedings of Machine Learning Research</i> 130:2242-2250 Available from https://proceedings.mlr.press/v130/kom-samo21a.html.
"""
import numpy as np

from kxy.misc.tf import CopulaLearner

def copula_entropy(z, subsets=[]):
	'''
	Estimate the entropy of the copula distribution of a d-dimensional random vector using MIND ([1]) with Spearman rank correlation constraints.


	Parameters
	----------
	z : np.array
		Vector whose rows are samples from the d-dimensional random vector and columns its coordinates.


	Returns
	-------
	ent : float
		The estimated copula entropy.
	'''
	if len(z.shape)==1 or z.shape[1]==1:
		return 0.0

	d = z.shape[1]
	cl = CopulaLearner(d, subsets=subsets)
	cl.fit(z)
	ent = min(cl.copula_entropy, 0.0)

	return ent



def mutual_information(y, x):
	'''
	Estimate the mutual information between two random vectors using MIND ([1]) with Spearman rank correlation constraints.


	Parameters
	----------
	y : np.array
		Vector whose rows are samples from the d-dimensional random vector and columns its coordinates.
	x : np.array
		Vector whose rows are samples from the d-dimensional random vector and columns its coordinates.


	Returns
	-------
	mi : float
		The estimated mutual information.
	'''
	y = y[:, None] if len(y.shape)==1 else y
	x = x[:, None] if len(x.shape)==1 else x
	z = np.concatenate([y, x], axis=1)
	huy = copula_entropy(y)
	hux = copula_entropy(x)
	huz = copula_entropy(z)
	mi = max(huy+hux-huz, 0.0)

	return mi


def run_d_dimensional_gaussian_experiment(d, rho, n=1000):
	'''
	'''
	# Cholesky decomposition of corr = np.array([[1., rho], [rho, 1.]])
	L = np.array([[1., 0.], [rho, np.sqrt(1.-rho*rho)]])
	y = np.empty((n, d))
	x = np.empty((n, d))
	for i in range(d):
		u = np.random.randn(n, 2)
		z = np.dot(L, u.T).T
		y[:, i] = z[:, 0].copy()
		x[:, i] = z[:, 1].copy()

	estimated_mi = mutual_information(y, x)
	true_mi = -d*0.5*np.log(1.-rho*rho)

	return estimated_mi, true_mi



if __name__ == '__main__':
	rho = 0.95
	d = 20
	estimated_mi, true_mi = run_d_dimensional_gaussian_experiment(d, rho)
	print('%dd Gaussian Mutual Information: Estimated %.4f, True (theoretical) %.4f' % (\
		d, estimated_mi, true_mi))



